/*!
** @file
** @brief   The implementation of the reticle used to find a target window.
** @remarks This feature is almost like the Finder Tool of Microsoft Spy++.
** @author  UETA, Shin-ichi <ueta@usdesign.jp>
** @license The MIT License
** @verbatim

Copyright (C) 2014 Ueta System Design Studio

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

@endverbatim
*/
#pragma once

/*!
** @brief   The base class of the site for the reticle.
** @remarks -
*/
class ATL_NO_VTABLE CReticleSite
{
public:

	/*!
	** @brief   Notifies the site that the reticle enters a top-level windows.
	** @param   [in] hWnd a handle to the top-level windows.
	** @returns TRUE to flash the window, or FALSE otherwise.
	** @remarks -
	*/
	virtual BOOL OnReticleEnter(HWND hWnd)
	{
		UNREFERENCED_PARAMETER(hWnd);

		return TRUE;
	}

	/*!
	** @brief   Notifies the site that the reticle leaves a top-level windows.
	** @param   [in] hWnd a handle to the top-level windows.
	** @returns No return value.
	** @remarks -
	*/
	virtual void OnReticleLeave(HWND hWnd)
	{
		UNREFERENCED_PARAMETER(hWnd);

		return;
	}

	/*!
	** @brief   Notifies the site that the reticle is losing the mouse capture.
	** @param   [in] hWnd a handle to the top-level windows.
	** @returns No return value.
	** @remarks -
	*/
	virtual void OnReticleRelease(HWND hWnd)
	{
		UNREFERENCED_PARAMETER(hWnd);

		return;
	}
};

/*!
** @brief   The reticle hosted by the Static control.
** @remarks -
*/
class CReticle : public CWindowImpl<CReticle, CStatic>
{
private:

	CReticleSite& m_site;

	CCursorHandle m_hPrevCursor;

	HWND m_hWndEnter;
	HWND m_hWndAimed;

public:

	/*!
	** @brief   Constructs a new CReticle object that is bound to the specified site.
	** @param   [in] site a reference to the site for the reticle.
	** @remarks -
	*/
	explicit CReticle(CReticleSite& site) : m_site(site)
	{
		m_hWndEnter = NULL;
		m_hWndAimed = NULL;
	}

	/*!
	** @brief   Subclasses a Static control and attaches it to the object.
	** @param   [in] hWnd a handle to the Static control to be subclassed.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL SubclassWindow(HWND hWnd)
	{
		if (!CWindowImpl<CReticle, CStatic>::SubclassWindow(hWnd))
			return FALSE;

		_U_STRINGorID cursor = GetDlgCtrlID();

		HCURSOR hCursor = AtlLoadCursorImage(cursor);
		ATLASSERT(hCursor);
		ATLASSERT(GetStyle() & SS_ICON);
		ATLVERIFY(SetIcon(HICON(hCursor)) == NULL);

		CWindow wndPrev = GetWindow(GW_HWNDPREV);
		ATLASSERT(wndPrev);
		ATLVERIFY(CenterWindow(wndPrev));

		return TRUE;
	}

	BEGIN_MSG_MAP(CReticle)
		MSG_WM_NCHITTEST(OnNCHitTest)
		MSG_WM_LBUTTONDOWN(OnLButtonDown)
		MSG_WM_LBUTTONUP(OnLButtonUp)
		MSG_WM_MOUSEMOVE(OnMouseMove)
		MSG_WM_CAPTURECHANGED(OnCaptureChanged)
	END_MSG_MAP()

	/*!
	** @brief   Handles the WM_NCHITTEST message.
	** @param   [in] point the position of the cursor, in screen coordinates.
	** @returns One of the mouse hit-test enumerated values.
	** @remarks -
	*/
	UINT OnNCHitTest(CPoint point)
	{
		UNREFERENCED_PARAMETER(point);

		return HTCLIENT;
	}

	/*!
	** @brief   Handles the WM_LBUTTONDOWN message.
	** @param   [in] nFlags indicates whether various virtual keys are down.
	** @param   [in] point the position of the cursor, in client-area coordinates.
	** @returns No return value.
	** @remarks -
	*/
	void OnLButtonDown(UINT nFlags, CPoint point)
	{
		UNREFERENCED_PARAMETER(nFlags);
		UNREFERENCED_PARAMETER(point);

		ICONINFO ii = {};

		HCURSOR hCursor = HCURSOR(GetIcon());
		ATLASSERT(hCursor);
		ATLVERIFY(::GetIconInfo(HICON(hCursor), &ii));
		CPoint ptHotspot(ii.xHotspot, ii.yHotspot);

		if (ii.hbmMask)
			ATLVERIFY(::DeleteObject(ii.hbmMask)), ii.hbmMask = NULL;
		if (ii.hbmColor)
			ATLVERIFY(::DeleteObject(ii.hbmColor)), ii.hbmColor = NULL;

		CRect rcReticle;
		ATLVERIFY(GetWindowRect(rcReticle));
		CPoint ptCursorPos = rcReticle.TopLeft() + ptHotspot;
		ATLVERIFY(::SetCursorPos(ptCursorPos.x, ptCursorPos.y));

		ShowWindow(SW_HIDE);

		m_hPrevCursor.Attach(::SetCursor(hCursor));

		SetCapture();
	}

	/*!
	** @brief   Handles the WM_LBUTTONUP message.
	** @param   [in] nFlags indicates whether various virtual keys are down.
	** @param   [in] point the position of the cursor, in client-area coordinates.
	** @returns No return value.
	** @remarks -
	*/
	void OnLButtonUp(UINT nFlags, CPoint point)
	{
		UNREFERENCED_PARAMETER(nFlags);

		if (::GetCapture() == m_hWnd)
		{
			::ReleaseCapture();

			CPoint ptCursorPos = point;
			ClientToScreen(&ptCursorPos);

			HWND hWndTarget = ::GetAncestor(::WindowFromPoint(ptCursorPos), GA_ROOT);

			m_site.OnReticleRelease(hWndTarget);
		}
	}

	/*!
	** @brief   Handles the WM_MOUSEMOVE message.
	** @param   [in] nFlags indicates whether various virtual keys are down.
	** @param   [in] point the position of the cursor, in client-area coordinates.
	** @returns No return value.
	** @remarks -
	*/
	void OnMouseMove(UINT nFlags, CPoint point)
	{
		UNREFERENCED_PARAMETER(nFlags);

		if (::GetCapture() == m_hWnd)
		{
			CPoint ptCursorPos = point;
			ClientToScreen(&ptCursorPos);

			HWND hWndTarget = ::GetAncestor(::WindowFromPoint(ptCursorPos), GA_ROOT);

			if (hWndTarget == m_hWndEnter) // hover
			{
				;
			}
			else
			{
				if (m_hWndEnter)
				{
					OnMouseLeave(m_hWndEnter);

					m_site.OnReticleLeave(m_hWndEnter);
				}
				m_hWndEnter = NULL;
				m_hWndAimed = NULL;

				if (hWndTarget)
				{
					m_hWndEnter = hWndTarget;

					if (m_site.OnReticleEnter(m_hWndEnter))
					{
						m_hWndAimed = m_hWndEnter;
					}
					OnMouseEnter(m_hWndEnter);
				}
			}
		}
	}

	/*!
	** @brief   Handles the WM_CAPTURECHANGED message.
	** @param   [in] hWnd a handle to the window gaining the mouse capture.
	** @returns No return value.
	** @remarks -
	*/
	void OnCaptureChanged(HWND hWnd)
	{
		UNREFERENCED_PARAMETER(hWnd);

		if (m_hWndEnter)
		{
			OnMouseLeave(m_hWndEnter);
		}
		m_hWndEnter = NULL;
		m_hWndAimed = NULL;

		::SetCursor(m_hPrevCursor.Detach());

		ShowWindow(SW_SHOW);
	}

	/*!
	** @brief   Called when the cursor enters a top-level windows.
	** @param   [in] hWnd a handle to the top-level windows.
	** @returns No return value.
	** @remarks -
	*/
	void OnMouseEnter(HWND hWnd)
	{
		if (hWnd == m_hWndAimed)
		{
			FLASHWINFO fwi = {sizeof(FLASHWINFO)};

			fwi.hwnd    = m_hWndAimed;
			fwi.dwFlags = FLASHW_TIMER | FLASHW_CAPTION;

			::FlashWindowEx(&fwi);
		}
	}

	/*!
	** @brief   Called when the cursor leaves a top-level windows.
	** @param   [in] hWnd a handle to the top-level windows.
	** @returns No return value.
	** @remarks -
	*/
	void OnMouseLeave(HWND hWnd)
	{
		if (hWnd == m_hWndAimed)
		{
			FLASHWINFO fwi = {sizeof(FLASHWINFO)};

			fwi.hwnd    = m_hWndAimed;
			fwi.dwFlags = FLASHW_STOP;

			::FlashWindowEx(&fwi);
		}
	}
};

/*!
** @brief   The implementation to host the reticle.
** @remarks -
*/
template <class T>
class ATL_NO_VTABLE CReticleSiteImpl : public CReticleSite
{
private:

	CReticle m_wndReticle;

public:

	/*!
	** @brief   Initializes a new CReticleSiteImpl object.
	** @remarks -
	*/
	CReticleSiteImpl() : m_wndReticle(*this)
	{
		;
	}

	/*!
	** @brief   Binds the site for the reticle to the owner.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL BindReticle()
	{
		T* pT = static_cast<T*>(this);

		CWindow wndCtrl = pT->GetDlgItem(IDC_RETICLE);
		ATLASSERT(wndCtrl);

		return m_wndReticle.SubclassWindow(wndCtrl);
	}
};
